import sys
import components

class SMT2(object):

	def __init__(self, format, logic, macro, verbose, filename):
		self.formula = []
		self.checking = ""
		self.last_aux = 1
		self.declarations = []
		self.shows = []
		self.format = format
		self.logic = logic
		self.macro = macro
		self.verbose = verbose
		self.filename = filename
		
	def commit(self):
		
		if self.format == 'smt2':
			FILE = open(self.filename,"w")
			if self.verbose == 1:
				FILE.write("""(set-option :produce-models true)
				""")
			FILE.write('(set-logic '+self.logic+')')
			FILE.write(self.macro)
			FILE.write("\n".join(self.declarations))
			FILE.write("\n\n")
			FILE.write("\n".join(self.formula))
			FILE.write("\n\n")
			FILE.write("""(check-sat)
			""")
			if self.verbose == 1:
				FILE.write("\n".join(self.shows))
			FILE.close()
		else:
			print '(benchmark foo'
			print ':logic '+self.logic
			print ':extrafuns (',
			print " ".join(self.declarations),
			print ' )'
			print ':formula'
			print '(and'
			print "\n".join(self.formula)
			print ')'
			print ')'

	def declare(self, var, type):
		if self.format == 'smt2':
			self.declarations.append('(declare-fun '+var+' () '+type+')')
		else:
			self.declarations.append('('+var+' '+type+')')
			
	def setInitial(self):
		self.header = ""
		if self.verbose == 1:
			self.header = '(set-option :produce-models true)'
		self.header = self.header+ '(set-logic '+self.logic+')'
	
	def box(self, var, values):
		if len(values) == 1:
			n=''
			fix = ''
		else:
			n=')'
			#values.sort(key=lambda x: x[0])
			fix = '(or '
		for i in values:
			if len(i) == 1:
				fix = fix + """
				(= """ + var + """ """ + i[0] + """) """
			else:
				fix = fix + """
				(and (>= """ + var + """ """ + i[0] + """) (<= """ + var + """ """ + i[1] + """)) """
				
		fix = fix + n
		self.assertion(fix)	
		
	def assertion(self, expr):
		if self.format == 'smt2':
			self.formula.append('(assert '+expr+' )')
		else:
			self.formula.append(expr)
			
	
	def ite(self, cond, x, y):
		return '(ite '+cond+' '+x+' '+y+')'

	def new_var(self, type):
		var = 'x'+str(self.last_aux)
		self.last_aux = self.last_aux + 1
		self.declare(var, type)
		return var
	
	def div(self, x, y, type):
		var = self.new_var(type)
		self.box(var,0,1)
		self.assertion(self.ite(self.bin('>',x,y),'(= (* '+x+' '+var+') '+y+')',self.bin('=',var,'1')))
		return var		
	
	def bin(self, op, x, y):
		return '('+op+' '+x+' '+y+')'
	
	def show(self, var):
		self.shows.append('(get-value((p1 '+var+')))')
		self.shows.append('(get-value((p2 '+var+')))')
		


class Ftnormlogic(SMT2):
	def __init__(self, componentes, verbose, FILE):
		theory = 'AUFLIRA'
		self.components = components.OrdinalSum(componentes)
		SMT2.__init__(self, 'smt2', theory, self.header(), verbose, FILE)

	def tnorm(self, x, y):
		return '(tnorm '+x+' '+y+')'
		
	def impl(self, x, y):
		return '(impl '+x+' '+y+')'
			
	def neg(self, x):
		return '(neg '+x+' '+y+')'
		               
	def dis(self, x, y):
		return '(dis '+x+' '+y+')'
	
	def con(self, x, y):
		return '(con '+x+' '+y+')'

	def pow(self, x, n):
		if n > 1:
			#return '(* '+x+' '+self.pow(x,n-1)+')'
			return self.tnorm(x, self.pow(x,n-1))
		else:
		    return x
	def powi(self, x, n):
		if n > 1:
			#return '(* '+x+' '+self.pow(x,n-1)+')'
		    return self.impl(auxx, self.powi(auxx,n-1))
		else:
		    return x
				
	def showVars(self, varlist):
		for var in varlist:
			self.show(var)
			
	def th(self, form):
		self.assertion("""(greater T """ + form + """)""")
	
	def val(self, form):
		self.assertion("(= "+ form + """ T)""")
		
	def affirm(self, eq):
		#mucho ojo con esto. Todo corre por cuenta del usuario, y debe tener en cuenta que manejo pares... Arreglar?
		self.assertion(eq)
		
	def header(self):
		h = """
		;Pair type definitions
		(define-sort Pair () (Array Int Real))
		(define-fun p1 ((x Pair)) Real
 		 (select x 1))
		(define-fun p2 ((x Pair)) Real
		 (select x 2))
		;truth const
		(declare-const T Pair)
		(assert (= T (store (store T 1 0.0) 2 (~ 1.0))))
		(declare-const F Pair)
		(assert (= F (store (store F 1 0.0) 2 0.0)))
		
		;min(x,y)
		(define-fun min ((x Pair) (y Pair)) Pair
		 (ite (= x T) y
		  (ite (= y T) x
		   (ite (< (p2 x) (p2 y)) x
		    (ite (< (p2 y) (p2 x)) y
		     (ite (<= (p1 x) (p1 y)) x y))))))	
		
		;max(x,y)
		(define-fun max ((x Pair) (y Pair)) Pair
		 (ite (= (min x y) x) y x))
						
		;conjunction min (x, y)
		(define-fun con ((x Pair) (y Pair)) Pair
		 (min x y))

		;disjunction max (x, y)
		(define-fun dis ((x Pair) (y Pair)) Pair
		 (max x y))
		
		;greater(x,y) T if x > y, F else
		(define-fun greater ((x Pair) (y Pair)) Bool
		 (ite (and (= x T) (> (p2 y) (~ 1.0))) true
		  (ite (> (p2 x) (p2 y)) true
		   (ite (and (= (p2 x) (p2 y)) (> (p1 x) (p1 y))) true
		    false))))

		;tnorm(x, y)
		(define-fun tnorm ((x Pair) (y Pair)) Pair
		 (ite (= (p2 x) (p2 y))
		 ;estan en la misma componente"""
		comps = self.components.getComponents()
		for c in comps:
			componentId = str(c.getId())+".0"
			h = h + """
			(ite (= (p2 x) """+componentId+ """)"""
			if c.getType() == 'l':
				if c.getNVals() == 0:
					maxim = '1.0'
				else:
					maxim = str(c.getNVals()-1)+".0"
				#minim + size*max{0, (x+y-size-2*minim)/size} = 0 + maxim* max{0,x+y-maxim/maxim} = max{0, x+y-maxim} 
				h = h + """
					(store x 1 (ite (>= (- (+ (p1 x) (p1 y)) """+maxim+""") 0.0) (- (+ (p1 x) (p1 y)) """+maxim+""") 0.0))"""
			elif c.getType() == 'g':
				h = h + """
					(store x 1 (ite (<= (p1 x) (p1 y)) (p1 x) (p1 y)))"""
			elif c.getType() == 'p':
				h = h + """
					(store x 1 (+ (p1 x) (p1 y)))"""
			else:
				#c.getType() == 'b' --> hay un unico elemento
				h = h + """
					x"""
		h = h + " x"
		for c in range(len(comps)):
			h = h + ')'
			
		h = h + """
			;estan en diferentes componentes
			(min x y)))
			
		;impl(x y) -residuum-
		(define-fun impl ((x Pair) (y Pair)) Pair
			(ite (or (= (min x y) x) (= x y)) 
				T
				(ite (or (> (p2 x) (p2 y)) (= x T)) y"""
		for c in comps:
			componentId = str(c.getId())+".0"
			h = h + """
			(ite (= (p2 x) """+componentId+ """)"""
			if c.getType() == 'l':
				if c.getNVals() == 0:
					maxim = '1.0'
				else:
					maxim = str(c.getNVals()-1)+".0"
				#minim + size*(1 - ((x-minim)/size) + ((y-minim)/size)) = minim + size - x+minim +y-minim = minim + size -x + y = size +y-x
				h = h + """
					(store x 1 (+ """+ maxim +""" (- (p1 y) (p1 x))))"""
			elif c.getType() == 'g':
				h = h + """
					y"""
			elif c.getType() == 'p':
				h = h + """
					(store x 1 (- (p1 y) (p1 x)))"""
			else:
				#nunca se dara este caso
				h = h + """
					x"""
		h = h + " x"
		for c in range(len(comps)):
			h = h + ')'
		h = h + """)))
		
		;negation (neg x = x -> 0)
		(define-fun neg ((x Pair)) Pair
		 (impl x F))"""
		
		return h
		
	def assignVars(self, variables):
		#for each group, do the proper assignments
		self.declarations.append("""
		
;variables definition
		""")
		comps = self.components.getComponents()		
		for var in variables:
			self.declare(var, 'Pair')
			self.declare(var+'i','Int')
			self.assertion("(< "+var+"i 0)")
			valoresP1 = ""
			#valores de la componente -por ahora, finitas componentes-
			valoresP2 = "(or (= "+var+" T) "
			for c in comps:
				valuec = str(c.getId())+".0"
				valoresP2 = valoresP2 + "(= (p2 "+var+") "+valuec+") "
				
				#valores de las variables como tales
				if len(comps) > 1:
					valoresP1 = valoresP1 + "(ite (= (p2 "+var+") "+valuec+") "
				
				ty = c.getType()
				nval = c.getNVals()
				if ty == 'l' or ty == 'g':
					if nval == 0:
						valoresP1 = valoresP1 + "(and (< (p1 "+var+") 1.0) (>= (p1 "+var+") 0.0)) "
					else:
						if nval > 2:
							valoresP1 = valoresP1 + "(or "
						for n in range(nval-1):
							valuev = str(n)+".0"
							valoresP1 = valoresP1 + "(= (p1 "+var+") "+valuev+") "
						if nval > 2:
							valoresP1 = valoresP1 + ") "
				else:
					#producto
					valoresP1 = valoresP1 + "(= (p1 "+var+") (to_real "+var+"i)) "
			if len(comps) > 1:
				valoresP1 = valoresP1 +"(= "+var+"i "+var+"i)"
				for c in comps:
					valoresP1 = valoresP1 + """)"""
			valoresP2 = valoresP2 + """)"""
			self.assertion(valoresP2)
			self.assertion(valoresP1)
			
	def assignConstants(self,constants):
		for c in constants:
			self.declarations.append('(declare-const '+c[0]+' Pair)')
			self.assertion("(= "+c[0]+" (store (store "+c[0]+" 1 "+c[1][0]+") 2 "+c[1][1]+".0))")