import sys

if __name__ == "__main__":
	import configuration
	import argparse
	import formparse
	import components
	import ftnorms
	import re
	import math
	import subprocess

	argsparser = argparse.ArgumentParser(prog='python fuzzy.py', 
		description='BL and extension logics reasoner.',
		formatter_class=argparse.RawTextHelpFormatter)
	argsparser.add_argument('-l', default='[l2]', help="""'bl' or a list [a+b+...] of components -logics- in {l, g, ln, gn, p}, n a natural number.""") 
#The Logic of the :
#	:bl --> (n+1)l, where n is the number of variables involved, 
#	:l --> [0,1] Lukasiewicz, 
#	:g --> [0,1] Godel, 
#	:ln, gn --> {0,...,n-1} Lukasiewicz/Godel (finite), 
#	:p --> l2 + Z-.""")
	argsparser.add_argument('-t', default='th', help="""Task over the logic: 
	th --> prove theoremhood of the formula in param. -f,
	c -->  prove consequence of the formula in -f from the premises in param. -p,
	s -->  assignment of a model that satisfies the set of equations in param. -eq.""")
	argsparser.add_argument('-f', default='T', help="""Formula in prefix notation. 
Constants as <a.b;c>, where a.b is the value (real)
in the component and c is the component id starting from 0.
T = true, F = false.""")
	argsparser.add_argument('-p', default='[]', help='List of formulas in prefix notation -possibly with constants-.')
	argsparser.add_argument('-eq', default='[]', help="""List of equations in the format (= formula formula') or (greater formula formula') 
-possibly with constants-.""")
	argsparser.add_argument('-out', default="output", help = "Name for the output file.")
	args = argsparser.parse_args()
	
	formsparser = formparse.FormParser()
	
	if args.l != 'bl':
		components = args.l[1:-1].split('+')
	
	outfile = "z3Codes/"+args.out	
	
	if args.t == 'th':
		#prove theoremhood
		variables = list(set(re.findall('(x[0-9]+(?=[ \)]))', args.f)))
		if args.l == 'bl':
			nvars = len(variables)
			components =  ["l" for i in range(len(variables) + 1)]
		#fix the constants in the formula
		#constants = list(set(re.findall('({.*?,[0-9]+})', args.f)))
		formsparser.addForm(args.f)
		nf = formsparser.parseForms()[0]
		logic = ftnorms.Ftnormlogic(components, 0, outfile)
		logic.assignVars(variables)
		logic.assignConstants(formsparser.getConstants())
		
		logic.th(nf)
		logic.commit()
			
	elif args.t == 'c':
		#prove logical consequence p --> f
		logic = ftnorms.Ftnormlogic(components, 0, outfile)
		premises = args.p[1:-1].split(',')
		variables = []
		for p in premises:
			variables.extend(list(set(re.findall('(x[0-9]+(?=[ \)]))', p))))
			formsparser.addForm(p)
		variables.extend(list(set(re.findall('(x[0-9]+(?=[ \)]))', args.f))))
		formsparser.addForm(args.f)

		variables = list(set(variables))

		nfs = formsparser.parseForms()
		npres = nfs[:-1]
		nf = nfs[len(nfs)-1]
		logic.assignVars(variables)
		logic.assignConstants(formsparser.getConstants())
		for p in npres:
			logic.val(p)
		logic.th(nf)
		
		logic.commit()
			
	elif args.t == 's':
		#assignment of a model satisfiying all equations in e
		logic = ftnorms.Ftnormlogic(components, 1, outfile)
		equations = args.eq[1:-1].split(',')
		variables = []
		for eq in equations:
			variables.extend(list(set(re.findall('(x[0-9]+(?=[ \)]))', eq))))
			formsparser.addForm(eq)
		variables = list(set(variables))
		
		neqs = formsparser.parseForms()
		logic.assignVars(variables)
		logic.assignConstants(formsparser.getConstants())
		for v in variables:
			logic.show(v)
		for e in neqs:
			logic.affirm(e)
		logic.commit()
		
		
	
	z3bin = configuration.Z3_LOCATION + "/bin/z3"
	outputz3 = subprocess.check_output([z3bin, "-smt2", outfile])
	init = list(set(re.findall('.*?(?=\s)', outputz3)))[1]
	
	if args.t == 'th':
		if init == 'unsat':
			print "Formula "+ args.f+" IS a theorem in logic "+ args.l
		else:
			print "Formula "+ args.f+" IS NOT a theorem in logic "+ args.l
	elif args.t == 'c':
		if init == 'unsat':
			print "Formula "+ args.f+" IS a logical consequence of " + args.p+" in logic "+ args.l
		else:
			print "Formula "+ args.f+" IS NOT a logical consequence of " + args.p+" in logic "+ args.l
	else:
		if init == 'unsat':
			print "The set of equations "+ args.eq+" DOES NOT have a model in logic "+ args.l
		else:
			print "The set of equations "+ args.eq+" DOES have a model in logic "+ args.l
			print "A possible assignation in the BL-chain assignated to "+ args.l +" of the involved variables is the following:"
			#processing of the output assigments
			output = outputz3.split('\n')[1:]
			i = 0
			while i < len(output)-1:
				p1 = re.findall('(x[0-9]+).*?((\(.*?\))|([0-9]+\.[0-9]+))', output[i])
				i = i+1
				p2 = re.findall('(x[0-9]+).*?(-)? ([0-9]+\.0)', output[i])
				i = i+1
				if p2[0][1] != "":
					value = 'T'
				else:
					if p1[0][1] == '0.0' and p2[0][2] =='0.0':
						value = 'F'
					else:
						if p1[0][1][0] == '(':
							l = p1[0][1][1:-1].split(" ")
							if l[0] == '/':
								val = float(l[1])/float(l[2])
							else:
								val = p1[0][1]
						else:
							val = p1[0][1]
						value = "<"+str(val)+";"+p2[0][2]+">"
				print p1[0][0]+" --> "+value